// File:     fir2.cc
// Author:   Jeff Miller, with code from Eric Wan
// Date:     12-19-97
// Abstract: 2-layer Finite Impulse Response neural network engine

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "firnet.h"

// Macros for readability and compatibility with original Eric Wan code
#define NUM_IN      (m_Nodes[0])
#define NUM_HID     (m_Nodes[1])
#define NUM_OUT     (m_Nodes[2])

void FIRNet2Layer::RandomizeWeights()
{
   Deallocate();

   m_WeightSize[0] = m_Nodes[0] * m_Nodes[1] * m_Taps[0];
   m_Weights[0] = new double[m_WeightSize[0]];
   m_LastWeights[0] = new double[m_WeightSize[0]];

   m_WeightSize[1] = m_Nodes[1] * m_Nodes[2] * m_Taps[1];
   m_Weights[1] = new double[m_WeightSize[1]];
   m_LastWeights[1] = new double[m_WeightSize[1]];

   m_Biases[0] = new double[m_Nodes[1]];
   m_LastBiases[0] = new double[m_Nodes[1]];
   m_Biases[1] = new double[m_Nodes[2]];
   m_LastBiases[1] = new double[m_Nodes[2]];

   assert(m_Weights[0] != NULL);
   assert(m_LastWeights[0] != NULL);
   assert(m_Weights[1] != NULL);
   assert(m_LastWeights[1] != NULL);
   assert(m_Biases[0] != NULL);
   assert(m_LastBiases[0] != NULL);
   assert(m_Biases[1] != NULL);
   assert(m_LastBiases[1] != NULL);

   RandomizeArray(m_Weights[0], m_WeightSize[0]);
   RandomizeArray(m_Weights[1], m_WeightSize[1]);
   RandomizeArray(m_Biases[0], m_Nodes[1]);
   RandomizeArray(m_Biases[1], m_Nodes[2]);

   ZeroArray(m_LastWeights[0], m_WeightSize[0]);
   ZeroArray(m_LastWeights[1], m_WeightSize[1]);
   ZeroArray(m_LastBiases[0], m_Nodes[1]);
   ZeroArray(m_LastBiases[1], m_Nodes[2]);
}

/********************************************************************/
/*                                                                  */
/* All of the following code is adapted from an original program by */
/* Eric Wan, Dept. of Computer Science, Oregon Graduate Institute   */
/*                                                                  */
/********************************************************************/

void FIRNet2Layer::Train(int NumVectors, double* InputVectors, double* TargetVectors)
{
   int    Epoch;
   int    i, j, tmp;
   double RateDecrease;

   m_NumInputVectors = NumVectors;
   m_InputVectors = InputVectors;
   m_TargetVectors = TargetVectors;

   // Create the circular index mask
   if (m_Taps[1] > 1)
      m_Mask = (1 << ((int)(log((double)(m_Taps[1] - 1)) / 0.69314718) + 1)) - 1;
   else
      m_Mask = 0;

   //
   // Allocate memory for data structures used only in training
   //

   m_HiddenOutput = new double*[NUM_HID];
   assert(m_HiddenOutput != NULL);
   for (i = 0; i < NUM_HID; i++)
      m_HiddenOutput[i] = new double[m_Mask + 1];

   m_NetOutput = new double[NUM_OUT];
   assert(m_NetOutput != NULL);

   m_HiddenDelta = new double[NUM_HID];
   assert(m_HiddenDelta != NULL);

   m_NetDelta = new double*[NUM_OUT];
   assert(m_NetDelta != NULL);
   for (i = 0; i < NUM_OUT; i++)
      m_NetDelta[i] = new double[m_Mask + 1];

   m_MSE = new double[NUM_OUT * m_NumEpochs];
   m_CSE = new double[NUM_OUT];
   assert(m_MSE != NULL);
   assert(m_CSE != NULL);

   if(m_DisplayFreq > 0)
   {
      printf("\nNodes: %i-%i-%i\n", NUM_IN, NUM_HID, NUM_OUT);
      printf("Taps: %i-%i\n", m_Taps[0], m_Taps[1]);
      printf("Training Vectors: %i\n", m_NumInputVectors);
      printf("Epochs: %li", m_NumEpochs);
      if (m_EpochsTrained > 0)
         printf(" (%i already trained)", m_EpochsTrained);
      printf("\nLearning Rate: %g\n", m_LearningRate);
      if (m_MomentumRate > 0.0)
         printf("Momentum Rate: %g\n", m_MomentumRate);
      printf("\n");
   }

   RateDecrease = m_LearningRate * 0.90 / (double)m_NumEpochs;

   for (Epoch = 0; Epoch < m_NumEpochs; Epoch++)
   {
      for (i = m_Taps[0] - 1; i < m_Taps[0] + m_Taps[1] - 2; i++)
         FwdPassLayer1(i);    // Prime the second layer taps

      for (; i < m_Taps[0] + 2 * m_Taps[1] - 3; i++)
      {
         FwdPassLayer1(i);
         FwdPassLayer2(i);
         BackPropLayer2(i);   // Prime the output layer deltas
         UpdateWeights(i);    // Adjust the second layer weights
      }

      // Train over the remaining input vectors
      for (; i < m_NumInputVectors; i++)
      {
         FwdPassLayer1(i);
         FwdPassLayer2(i);
         BackPropLayer2(i);
         BackPropLayer1(i);
         UpdateWeights(i);
      }

      // Complete backprop of deltas
      for (; i < m_NumInputVectors + m_Taps[1] - 1; i++)
      {
         for (tmp = i & m_Mask, j = 0; j < NUM_OUT; j++)
            m_NetDelta[j][tmp] = 0.0;

         BackPropLayer1(i);
         UpdateWeights(i);
      }

      CalcMSE(Epoch);
      ResetDeltas();
      m_EpochsTrained++;

      if (m_DecreaseRate)
         m_LearningRate -= RateDecrease;
   }

   if ((m_NumEpochs % m_DisplayFreq) != 0.0 && m_DisplayFreq > 0)
   {
      printf("Epoch: %li MSE = ", m_NumEpochs);
      for (i = 0; i < NUM_OUT; i++)
         printf(" %g", m_MSE[COL_ROW(i, m_NumEpochs - 1, NUM_OUT)]);

      if (m_DecreaseRate)
         printf(" Rate = %g", m_LearningRate);

      printf("\n");
   }

   //
   // Deallocate training structures
   //

   for (i = 0; i < NUM_HID; i++)
      delete [] m_HiddenOutput[i];
   delete [] m_HiddenOutput;

   delete [] m_NetOutput;
   delete [] m_HiddenDelta;

   for (i = 0; i < NUM_OUT; i++)
      delete [] m_NetDelta[i];
   delete [] m_NetDelta;

   delete [] m_MSE;
   delete [] m_CSE;
}

void FIRNet2Layer::FwdPassLayer1(int Index)
{
   int    j, k, t;
   int    tmp, tmp1;
   int    RowsByCols = NUM_HID * NUM_IN;
   double Out;

   tmp1 = Index & m_Mask;

   for (k = 0; k < NUM_HID; k++)       // For each node in the hidden layer
   {
      Out = m_Biases[0][k];
      for (t = 0; t < m_Taps[0]; t++)  // For each tap in the hidden layer
      {
         tmp = Index - t;
         for (j = 0; j < NUM_IN; j++)  // Weighted sum of inputs
            Out += m_InputVectors[COL_ROW(j, tmp, NUM_IN)] *
               m_Weights[0][COL_ROW_TAP(k, j, t, NUM_HID, RowsByCols)];
      }
      m_HiddenOutput[k][tmp1] = HID_FN(Out); // Transfer function
   }
}

void FIRNet2Layer::FwdPassLayer2(int Index)
{
   int    j, k, t;
   int    tmp;
   double Out;
   int    RowsByCols = NUM_OUT * NUM_HID;

   for (k = 0; k < NUM_OUT; k++)       // For node in the output layer
   {
      Out = m_Biases[1][k];
      for (t = 0; t < m_Taps[1]; t++)  // For each time tap in the output layer
      {
         tmp = (Index - t) & m_Mask;
         for (j = 0; j < NUM_HID; j++) // Inner product
            Out += m_HiddenOutput[j][tmp] *
               m_Weights[1][COL_ROW_TAP(k, j, t, NUM_OUT, RowsByCols)];
      }
      m_NetOutput[k] = OUT_FN(Out);    // Transfer function
   }
}

void FIRNet2Layer::BackPropLayer1(int Index)
{
   int    i, j, t, tmp;
   int    RowsByCols = NUM_OUT * NUM_HID;
   double Delta;

   for (i = 0; i < NUM_HID; i++)       // For each node in the hidden layer
   {
      Delta = 0.0;
      for (t = 0; t < m_Taps[1]; t++)  // For each time tap in the output layer
      {
         tmp = (Index - t) & m_Mask;
         for (j = 0; j < NUM_OUT; j++)
            Delta += m_NetDelta[j][tmp] *
               m_Weights[1][COL_ROW_TAP(j, i, m_Taps[1] - 1 - t, NUM_OUT, RowsByCols)];
      }
      tmp = (Index - m_Taps[1] + 1) & m_Mask;
      m_HiddenDelta[i] = Delta * D_HID_FN(m_HiddenOutput[i][tmp]);
   }
}

void FIRNet2Layer::BackPropLayer2(int Index)
{
   int    i, tmp;
   double Err;

   tmp = Index & m_Mask;

   for (i = 0; i < NUM_OUT; i++)
   {
      // Compute the error between the computed and desired output
      Err = m_TargetVectors[COL_ROW(i, Index, NUM_OUT)] - m_NetOutput[i];
      m_CSE[i] += Err * Err;
      m_NetDelta[i][tmp] = Err;
   }
}

void FIRNet2Layer::UpdateWeights(int Index)
{
   int    i, j, t;
   int    tmp;
   int    RowsByCols = NUM_OUT * NUM_HID;
   double Delta, Change;

   for (i = 0; i < NUM_OUT; i++) // Output layer weights
   {
      Delta = m_LearningRate * m_NetDelta[i][Index & m_Mask];
      for (t = 0; t < m_Taps[1]; t++)
      {
         tmp = (Index - t) & m_Mask;
         for (j = 0; j < NUM_HID; j++)
         {
            Change = Delta * m_HiddenOutput[j][tmp] +
               m_MomentumRate * m_LastWeights[1][COL_ROW_TAP(i, j, t, NUM_OUT, RowsByCols)];

            m_Weights[1][COL_ROW_TAP(i, j, t, NUM_OUT, RowsByCols)] += Change;
            m_LastWeights[1][COL_ROW_TAP(i, j, t, NUM_OUT, RowsByCols)] = Change;
         }
      }
      Change = Delta + m_MomentumRate * m_LastBiases[1][i];
      m_Biases[1][i] += Change;
      m_LastBiases[1][i] = Change;
   }

   RowsByCols = NUM_IN * NUM_HID;

   for (i = 0; i < NUM_HID; i++) // Hidden layer weights
   {
      Delta = m_LearningRate * m_HiddenDelta[i];
      for (t = 0; t < m_Taps[0]; t++)
      {
         tmp = Index - t - m_Taps[1] + 1;
         for (j = 0; j < NUM_IN; j++)
         {
            Change = Delta * m_InputVectors[COL_ROW(j, tmp, NUM_IN)] +
               m_MomentumRate * m_LastWeights[0][COL_ROW_TAP(i, j, t, NUM_HID, RowsByCols)];
            m_Weights[0][COL_ROW_TAP(i, j, t, NUM_HID, RowsByCols)] += Change;
            m_LastWeights[0][COL_ROW_TAP(i, j, t, NUM_HID, RowsByCols)] = Change;
         }
      }
      Change = Delta + m_MomentumRate * m_LastBiases[0][i];
      m_Biases[0][i] += Change;
      m_LastBiases[0][i] = Change;
   }
}

void FIRNet2Layer::CalcMSE(int Epoch)
{
   int i;

   for (i = 0; i < NUM_OUT; i++)
   {
      m_MSE[COL_ROW(i, Epoch, NUM_OUT)] =
         m_CSE[i] / (double)(m_NumInputVectors - m_Taps[1] - m_Taps[0] + 2);
      m_CSE[i] = 0.0;
   }

   if (((Epoch + 1) % m_DisplayFreq) == 0)
   {
      printf("Epoch: %li MSE = ", Epoch + 1);
      for (i = 0; i < NUM_OUT; i++)
         printf(" %g", m_MSE[COL_ROW(i, Epoch, NUM_OUT)]);

      if (m_DecreaseRate)
         printf(" Rate = %g", m_LearningRate);

      printf("\n");
   }
}

void FIRNet2Layer::ResetDeltas()
{
   int i, t;

   for (i = 0; i < NUM_OUT; i++)
   {
      for (t = m_Taps[1]; t < m_Mask + 1; t++)
         m_NetDelta[i][t] = 0.0;
   }
}

void FIRNet2Layer::ForwardPass()
{
   int     i;
   double* OrigOutput;

   // Create the circular index mask
   if (m_Taps[1] > 1)
      m_Mask = (1 << ((int)(log((double)(m_Taps[1] - 1)) / 0.69314718) + 1)) - 1;
   else
      m_Mask = 0;

   m_HiddenOutput = new double*[NUM_HID];
   assert(m_HiddenOutput != NULL);
   for (i = 0; i < NUM_HID; i++)
      m_HiddenOutput[i] = new double[m_Mask + 1];

   for (i = m_Taps[0] - 1; i < m_Taps[0] + m_Taps[1] - 2; i++)
      FwdPassLayer1(i);    // Prime the second layer of taps

   OrigOutput = m_NetOutput;
   m_NetOutput += i * NUM_OUT;   // Skip past primer values

   // Go over all remaining input vectors
   for (; i < m_NumInputVectors; i++)
   {
      FwdPassLayer1(i);
      FwdPassLayer2(i);    // Output goes into m_NetOutput
      m_NetOutput += NUM_OUT;
   }

   // There should be (NumInputVectors - WindowSize) output points

   m_NetOutput = OrigOutput;

   for (i = 0; i < NUM_HID; i++)
      delete [] m_HiddenOutput[i];
   delete [] m_HiddenOutput;
}
