// File:     fir3.cc
// Author:   Jeff Miller, with code from Eric Wan
// Date:     12-19-97
// Abstract: 3-layer Finite Impulse Response neural network engine

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "firnet.h"

// Macros for readability and compatibility with original Eric Wan code
#define NUM_IN      (m_Nodes[0])
#define NUM_HID1    (m_Nodes[1])
#define NUM_HID2    (m_Nodes[2])
#define NUM_OUT     (m_Nodes[3])

void FIRNet3Layer::RandomizeWeights()
{
   Deallocate();

   m_WeightSize[0] = m_Nodes[0] * m_Nodes[1] * m_Taps[0];
   m_Weights[0] = new double[m_WeightSize[0]];
   m_LastWeights[0] = new double[m_WeightSize[0]];

   m_WeightSize[1] = m_Nodes[1] * m_Nodes[2] * m_Taps[1];
   m_Weights[1] = new double[m_WeightSize[1]];
   m_LastWeights[1] = new double[m_WeightSize[1]];

   m_WeightSize[2] = m_Nodes[2] * m_Nodes[3] * m_Taps[2];
   m_Weights[2] = new double[m_WeightSize[2]];
   m_LastWeights[2] = new double[m_WeightSize[2]];

   m_Biases[0] = new double[m_Nodes[1]];
   m_LastBiases[0] = new double[m_Nodes[1]];
   m_Biases[1] = new double[m_Nodes[2]];
   m_LastBiases[1] = new double[m_Nodes[2]];
   m_Biases[2] = new double[m_Nodes[3]];
   m_LastBiases[2] = new double[m_Nodes[3]];

   assert(m_Weights[0] != NULL);
   assert(m_LastWeights[0] != NULL);
   assert(m_Weights[1] != NULL);
   assert(m_LastWeights[1] != NULL);
   assert(m_Weights[2] != NULL);
   assert(m_LastWeights[2] != NULL);
   assert(m_Biases[0] != NULL);
   assert(m_LastBiases[0] != NULL);
   assert(m_Biases[1] != NULL);
   assert(m_LastBiases[1] != NULL);
   assert(m_Biases[2] != NULL);
   assert(m_LastBiases[2] != NULL);

   RandomizeArray(m_Weights[0], m_WeightSize[0]);
   RandomizeArray(m_Weights[1], m_WeightSize[1]);
   RandomizeArray(m_Weights[2], m_WeightSize[2]);
   RandomizeArray(m_Biases[0], m_Nodes[1]);
   RandomizeArray(m_Biases[1], m_Nodes[2]);
   RandomizeArray(m_Biases[2], m_Nodes[3]);

   ZeroArray(m_LastWeights[0], m_WeightSize[0]);
   ZeroArray(m_LastWeights[1], m_WeightSize[1]);
   ZeroArray(m_LastWeights[2], m_WeightSize[2]);
   ZeroArray(m_LastBiases[0], m_Nodes[1]);
   ZeroArray(m_LastBiases[1], m_Nodes[2]);
   ZeroArray(m_LastBiases[2], m_Nodes[3]);
}

/********************************************************************/
/*                                                                  */
/* All of the following code is adapted from an original program by */
/* Eric Wan, Dept. of Computer Science, Oregon Graduate Institute   */
/*                                                                  */
/********************************************************************/

void FIRNet3Layer::Train(int NumVectors, double* InputVectors, double* TargetVectors)
{
   int    Epoch;
   int    i, j, tmp;
   double RateDecrease;

   m_NumInputVectors = NumVectors;
   m_InputVectors = InputVectors;
   m_TargetVectors = TargetVectors;

   // Create the circular index masks
   if (m_Taps[1] + m_Taps[2] > 2)
      m_Mask = (1 << ((int)(log((double)(m_Taps[1] + m_Taps[2] - 2)) / 0.69314718) + 1)) - 1;
   else
      m_Mask = 0;

   if (m_Taps[2] > 1)
      m_Mask2 = (1 << ((int)(log((double)(m_Taps[2] - 1)) / 0.69314718) + 1)) - 1;
   else
      m_Mask2 = 0;

   //
   // Allocate memory for data structures used only in training
   //

   m_HiddenOutput = new double*[NUM_HID1];
   assert(m_HiddenOutput != NULL);
   for (i = 0; i < NUM_HID1; i++)
      m_HiddenOutput[i] = new double[m_Mask + 1];

   m_HiddenOutput2 = new double*[NUM_HID2];
   assert(m_HiddenOutput2 != NULL);
   for (i = 0; i < NUM_HID2; i++)
      m_HiddenOutput2[i] = new double[m_Mask2 + 1];

   m_NetOutput = new double[NUM_OUT];
   assert(m_NetOutput != NULL);

   m_HiddenDelta = new double[NUM_HID1];
   assert(m_HiddenDelta != NULL);

   m_HiddenDelta2 = new double*[NUM_HID2];
   assert(m_HiddenDelta2 != NULL);
   for (i = 0; i < NUM_HID2; i++)
      m_HiddenDelta2[i] = new double[m_Mask + 1];

   m_NetDelta = new double*[NUM_OUT];
   assert(m_NetDelta != NULL);
   for (i = 0; i < NUM_OUT; i++)
      m_NetDelta[i] = new double[m_Mask2 + 1];

   m_MSE = new double[NUM_OUT * m_NumEpochs];
   m_CSE = new double[NUM_OUT];

   assert(m_MSE != NULL);
   assert(m_CSE != NULL);

   if (m_DisplayFreq > 0)
   {
      printf("\nNodes: %i-%i-%i-%i\n", NUM_IN, NUM_HID1, NUM_HID2, NUM_OUT);
      printf("Taps: %i-%i-%i\n", m_Taps[0], m_Taps[1], m_Taps[2]);
      printf("Training Vectors: %i\n", m_NumInputVectors);
      printf("Epochs: %li", m_NumEpochs);
      if (m_EpochsTrained > 0)
         printf(" (%i already trained)", m_EpochsTrained);
      printf("\nLearning Rate: %g\n", m_LearningRate);
      if (m_MomentumRate > 0.0)
         printf("Momentum Rate: %g\n", m_MomentumRate);
      printf("\n");
   }

   RateDecrease = m_LearningRate * 0.90 / (double)m_NumEpochs;

   for (Epoch = 0; Epoch < m_NumEpochs; Epoch++)
   {
      for (i = m_Taps[0] - 1; i < m_Taps[0] + m_Taps[1] - 2; i++)
         FwdPassLayer1(i);    // Prime the second layer taps

      for (; i < m_Taps[0] + m_Taps[1] + m_Taps[2] - 3; i++)
      {
         FwdPassLayer1(i);    // Prime the third layer taps
         FwdPassLayer2(i);
      }

      for (; i < m_Taps[0] + m_Taps[1] + 2 * m_Taps[2] - 4; i++)
      {
         FwdPassLayer1(i);    // Prime the output layer deltas
         FwdPassLayer2(i);
         FwdPassLayer3(i);
         BackPropLayer3(i);
         UpdateWeights(i);    // Adjust third layer weights
      }

      for (; i< m_Taps[0] + 2 * m_Taps[1] + 2 * m_Taps[2] - 5; i++)
      {
         FwdPassLayer1(i);    // Prime the second hidden layer deltas
         FwdPassLayer2(i);
         FwdPassLayer3(i);
         BackPropLayer3(i);
         BackPropLayer2(i);
         UpdateWeights(i);    // Adjust both hidden layer weights
      }

      // Train over the remaining input vectors
      for (; i < m_NumInputVectors; i++)
      {
         FwdPassLayer1(i);
         FwdPassLayer2(i);
         FwdPassLayer3(i);
         BackPropLayer3(i);
         BackPropLayer2(i);
         BackPropLayer1(i);
         UpdateWeights(i);
      }

      // Complete backprop of deltas
      for (; i < m_NumInputVectors + m_Taps[2] + m_Taps[1] - 2; i++)
      {
         tmp = i & m_Mask2;
         for (j = 0; j < NUM_OUT; j++)
            m_NetDelta[j][tmp] = 0.0;

         BackPropLayer2(i);
         BackPropLayer1(i);
         UpdateWeights(i);
      }

      ResetDeltas();
      CalcMSE(Epoch);
      m_EpochsTrained++;

      if (m_DecreaseRate)
         m_LearningRate -= RateDecrease;
   }

   if ((m_NumEpochs % m_DisplayFreq) != 0.0 && m_DisplayFreq > 0)
   {
      printf("Epoch: %li MSE = ", m_NumEpochs);
      for (i = 0; i < NUM_OUT; i++)
         printf(" %g", m_MSE[COL_ROW(i, m_NumEpochs - 1, NUM_OUT)]);

      if (m_DecreaseRate)
         printf(" Rate = %g", m_LearningRate);

      printf("\n");
   }

   //
   // Deallocate training structures
   //

   for (i = 0; i < NUM_HID1; i++)
      delete [] m_HiddenOutput[i];
   delete [] m_HiddenOutput;

   for (i = 0; i < NUM_HID2; i++)
      delete [] m_HiddenOutput2[i];
   delete [] m_HiddenOutput2;

   delete [] m_NetOutput;
   delete [] m_HiddenDelta;

   for (i = 0; i < NUM_HID2; i++)
      delete [] m_HiddenDelta2[i];
   delete [] m_HiddenDelta2;

   for (i = 0; i < NUM_OUT; i++)
      delete [] m_NetDelta[i];
   delete [] m_NetDelta;

   delete [] m_MSE;
   delete [] m_CSE;
}

void FIRNet3Layer::FwdPassLayer1(int Index)
{
   int    j, k, t;
   int    tmp, tmp1;
   int    RowsByCols = NUM_HID1 * NUM_IN;
   double Out;

   tmp1 = Index & m_Mask;
   for (k = 0; k < NUM_HID1; k++)   // For each node in the 1st hidden layer
   {
      Out = m_Biases[0][k];
      for (t = 0; t < m_Taps[0]; t++)  // For each tap in the 1st hidden layer
      {
         tmp = Index - t;
         for (j = 0; j < NUM_IN; j++)  // Weighted sum of inputs
            Out += m_InputVectors[COL_ROW(j, tmp, NUM_IN)] *
               m_Weights[0][COL_ROW_TAP(k, j, t, NUM_HID1, RowsByCols)];
      }
      m_HiddenOutput[k][tmp1] = HID_FN(Out); // Transfer function
   }
}

void FIRNet3Layer::FwdPassLayer2(int Index)
{
   int    j, k, t;
   int    tmp, tmp1;
   int    RowsByCols = NUM_HID2 * NUM_HID1;
   double Out;

   tmp1 = Index & m_Mask2;

   for (k = 0; k < NUM_HID2; k++)   // For each node in the 2nd hidden layer
   {
      Out = m_Biases[1][k];
      for (t = 0; t < m_Taps[1]; t++)  // For each tap in the 2nd hidden layer
      {
         tmp = (Index - t) & m_Mask;
         for (j = 0; j < NUM_HID1; j++)   // Inner product
         Out += m_HiddenOutput[j][tmp] *
            m_Weights[1][COL_ROW_TAP(k, j, t, NUM_HID2, RowsByCols)];
      }
      m_HiddenOutput2[k][tmp1] = HID_FN(Out);   // Transfer function
   }
}

void FIRNet3Layer::FwdPassLayer3(int Index)
{
   int    j, k, t;
   int    tmp;
   int    RowsByCols = NUM_OUT * NUM_HID2;
   double Out;

   for (k = 0; k < NUM_OUT; k++)    // For each node in the output layer
   {
      Out = m_Biases[2][k];
      for (t = 0; t < m_Taps[2]; t++)  // For each tap in the output layer
      {
         tmp = (Index - t) & m_Mask2;
         for (j = 0; j < NUM_HID2; j++)   // Inner product
            Out += m_HiddenOutput2[j][tmp] *
               m_Weights[2][COL_ROW_TAP(k, j, t, NUM_OUT, RowsByCols)];
      }
      m_NetOutput[k] = OUT_FN(Out);
   }
}

void FIRNet3Layer::BackPropLayer3(int Index)
{
   int    i, tmp;
   double Err;

   tmp = Index & m_Mask2;

   for (i = 0; i < NUM_OUT; i++)
   {
      // Compute the error between the computed and desired output
      Err = m_TargetVectors[COL_ROW(i, Index, NUM_OUT)] - m_NetOutput[i];
      m_CSE[i] += Err * Err;
      m_NetDelta[i][tmp] = Err;
   }
}

void FIRNet3Layer::BackPropLayer2(int Index)
{
   int    i, j, t;
   int    tmp, tmp1;
   int    RowsByCols = NUM_OUT * NUM_HID2;
   double Delta;

   tmp1 = (Index - m_Taps[2] + 1) & m_Mask2;

   for (i = 0; i < NUM_HID2; i++)   // For each node in the 2nd hidden layer
   {
      Delta = 0.0;
      for (t = 0; t < m_Taps[2]; t++)  // For each tap in the 2nd hidden layer
      {
         tmp = (Index - t) & m_Mask2;
         for (j = 0; j < NUM_OUT; j++)
         Delta += m_NetDelta[j][tmp] *
            m_Weights[2][COL_ROW_TAP(j, i, m_Taps[2] - 1 - t, NUM_OUT, RowsByCols)];
      }
      m_HiddenDelta2[i][Index & m_Mask] = Delta * D_HID_FN(m_HiddenOutput2[i][tmp1]);
   }
}

void FIRNet3Layer::BackPropLayer1(int Index)
{
   int    i, j, t;
   int    tmp, tmp2;
   int    RowsByCols = NUM_HID2 * NUM_HID1;
   double Delta;

   tmp2 = (Index - m_Taps[1] - m_Taps[2] + 2) & m_Mask;

   for (i = 0; i < NUM_HID1; i++)   // For each node in the 1st hidden layer
   {
      Delta = 0.0;
      for (t = 0; t < m_Taps[1]; t++)  // For each tap in the 1st hidden layer
      {
         tmp = (Index - t) & m_Mask;
         for (j = 0; j < NUM_HID2; j++)
         Delta += m_HiddenDelta2[j][tmp] *
            m_Weights[1][COL_ROW_TAP(j, i, m_Taps[1] - 1 - t, NUM_HID2, RowsByCols)];
      }
      m_HiddenDelta[i] = Delta * D_HID_FN(m_HiddenOutput[i][tmp2]);
   }
}

void FIRNet3Layer::UpdateWeights(int Index)
{
   int    i, j, t;
   int    tmp, tmp1;
   int    RowsByCols = NUM_OUT * NUM_HID2;
   double Delta, Change;

   for (i = 0; i < NUM_OUT; i++)    // Output layer weights
   {
      Delta = m_LearningRate * m_NetDelta[i][Index & m_Mask2];
      for (t = 0; t < m_Taps[2]; t++)
      {
         tmp = (Index - t) & m_Mask2;
         for (j = 0; j < NUM_HID2; j++)
         {
            Change = Delta * m_HiddenOutput2[j][tmp] +
               m_MomentumRate * m_LastWeights[2][COL_ROW_TAP(i, j, t, NUM_OUT, RowsByCols)];
            m_Weights[2][COL_ROW_TAP(i, j, t, NUM_OUT, RowsByCols)] += Change;
            m_LastWeights[2][COL_ROW_TAP(i, j, t, NUM_OUT, RowsByCols)] = Change;
         }
      }
      Change = Delta + m_MomentumRate * m_LastBiases[2][i];
      m_Biases[2][i] += Change;
      m_LastBiases[2][i] = Change;
   }

   RowsByCols = NUM_HID2 * NUM_HID1;

   tmp1 = Index - m_Taps[2] + 1;
   for (i = 0; i < NUM_HID2; i++)   // 2nd hidden layer weights
   {
      Delta = m_LearningRate * m_HiddenDelta2[i][Index & m_Mask];
      for (t = 0; t < m_Taps[1]; t++)
      {
         tmp = (tmp1 - t) & m_Mask;
         for (j = 0; j < NUM_HID1; j++)
         {
            Change = Delta * m_HiddenOutput[j][tmp] +
               m_MomentumRate * m_LastWeights[1][COL_ROW_TAP(i, j, t, NUM_HID2, RowsByCols)];
            m_Weights[1][COL_ROW_TAP(i, j, t, NUM_HID2, RowsByCols)] += Change;
            m_LastWeights[1][COL_ROW_TAP(i, j, t, NUM_HID2, RowsByCols)] = Change;
         }
      }
      Change = Delta + m_MomentumRate * m_LastBiases[1][i];
      m_Biases[1][i] += Change;
      m_LastBiases[1][i] = Change;
   }

   RowsByCols = NUM_HID1 * NUM_IN;

   tmp1 = Index - m_Taps[1] - m_Taps[2] + 2;
   for (i = 0; i < NUM_HID1; i++)   // 1st hidden layer weights
   {
      Delta = m_LearningRate * m_HiddenDelta[i];
      for (t = 0; t < m_Taps[0]; t++)
      {
         tmp = tmp1 - t;
         for (j = 0; j < NUM_IN; j++)
         {
            Change = Delta * m_InputVectors[COL_ROW(j, tmp, NUM_IN)] +
               m_MomentumRate * m_LastWeights[0][COL_ROW_TAP(i, j, t, NUM_HID1, RowsByCols)];
            m_Weights[0][COL_ROW_TAP(i, j, t, NUM_HID1, RowsByCols)] += Change;
            m_LastWeights[0][COL_ROW_TAP(i, j, t, NUM_HID1, RowsByCols)] = Change;
         }
      }
      Change = Delta + m_MomentumRate * m_LastBiases[0][i];
      m_Biases[0][i] += Change;
      m_LastBiases[0][i] = Change;
   }
}

void FIRNet3Layer::CalcMSE(int Epoch)
{
   int i;

   for (i = 0; i < NUM_OUT; i++)
   {
      m_MSE[COL_ROW(i, Epoch, NUM_OUT)] =
         m_CSE[i] / (double)(m_NumInputVectors - m_Taps[0] - m_Taps[1] - m_Taps[2] + 3);
      m_CSE[i] = 0.0;
   }

   if (((Epoch + 1) % m_DisplayFreq) == 0)
   {
      printf("Epoch: %li MSE = ", Epoch + 1);
      for (i = 0; i < NUM_OUT; i++)
         printf(" %g", m_MSE[COL_ROW(i, Epoch, NUM_OUT)]);

      if (m_DecreaseRate)
         printf(" Rate = %g", m_LearningRate);

      printf("\n");
   }
}

void FIRNet3Layer::ResetDeltas()
{
   int i, t;

   for (i = 0; i < NUM_OUT; i++)
   {
      for (t = m_Taps[1] + m_Taps[2]; t < m_Mask2 + 1; t++)
         m_NetDelta[i][t] = 0.0;
   }

   for (i = 0; i < NUM_HID2; i++)
   {
      for (t = m_Taps[1]; t <= m_Mask; t++)
         m_HiddenDelta2[i][t] = 0.0;
   }
}

void FIRNet3Layer::ForwardPass(void)
{
   int     i;
   double* OrigOutput;

   //
   // Create the circular index masks
   //
   if (m_Taps[1] + m_Taps[2] > 2)
      m_Mask = (1 << ((int)(log((double)(m_Taps[1] + m_Taps[2] - 2)) / 0.69314718) + 1)) - 1;
   else
      m_Mask = 0;

   if (m_Taps[2] > 1)
      m_Mask2 = (1 << ((int)(log((double)(m_Taps[2] - 1)) / 0.69314718) + 1)) - 1;
   else
      m_Mask2 = 0;

   m_HiddenOutput = new double*[NUM_HID1];
   assert(m_HiddenOutput != NULL);
   for (i = 0; i < NUM_HID1; i++)
      m_HiddenOutput[i] = new double[m_Mask + 1];

   m_HiddenOutput2 = new double*[NUM_HID2];
   assert(m_HiddenOutput2 != NULL);
   for (i = 0; i < NUM_HID2; i++)
      m_HiddenOutput2[i] = new double[m_Mask2 + 1];

   for (i = m_Taps[0] - 1; i < m_Taps[0] + m_Taps[1] - 2; i++)
      FwdPassLayer1(i);    // Prime the second layer of taps

   for (; i < m_Taps[0] + m_Taps[1] + m_Taps[2] - 3; i++)
   {
      FwdPassLayer1(i);    // Prime the third layer of taps
      FwdPassLayer2(i);
   }

   OrigOutput = m_NetOutput;
   m_NetOutput += i * NUM_OUT;   // Skip past primer values

   // Go over all remaining input vectors
   for (; i < m_NumInputVectors; i++)
   {
      FwdPassLayer1(i);
      FwdPassLayer2(i);
      FwdPassLayer3(i);
      m_NetOutput += NUM_OUT;
   }

   // There should be (NumInputVectors - WindowSize) output points

   m_NetOutput = OrigOutput;

   for (i = 0; i < NUM_HID1; i++)
      delete [] m_HiddenOutput[i];
   delete [] m_HiddenOutput;

   for (i = 0; i < NUM_HID2; i++)
      delete [] m_HiddenOutput2[i];
   delete [] m_HiddenOutput2;
}
