// File:     firnet.cc
// Author:   Jeff Miller
// Date:     12-18-97
// Abstract: Generic 2 or 3-layer Finite Impulse Response neural network engine

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "firnet.h"

FIRNet::FIRNet()
{
   m_Weights[0] = m_Weights[1] = m_Weights[2] = NULL;
   m_Biases[0] = m_Biases[1] = m_Biases[2] = NULL;
   m_LastWeights[0] = m_LastWeights[1] = m_LastWeights[2] = NULL;
   m_LastBiases[0] = m_LastBiases[1] = m_LastBiases[2] = NULL;
   m_EpochsTrained = 0;
}

FIRNet& FIRNet::operator=(FIRNet& net)
{
   int i;

   for (i = 0; i < 4; i++)
      m_Nodes[i] = net.m_Nodes[i];

   for (i = 0; i < 3; i++)
      m_Taps[i] = net.m_Taps[i];

   return *this;
}

void FIRNet::Deallocate()
{
   // Massive memory deallocation
   if (m_Weights[0] != NULL) delete [] m_Weights[0];
   if (m_Weights[1] != NULL) delete [] m_Weights[1];
   if (m_Weights[2] != NULL) delete [] m_Weights[2];
   if (m_Biases[0] != NULL) delete [] m_Biases[0];
   if (m_Biases[1] != NULL) delete [] m_Biases[1];
   if (m_Biases[2] != NULL) delete [] m_Biases[2];
   if (m_LastWeights[0] != NULL) delete [] m_LastWeights[0];
   if (m_LastWeights[1] != NULL) delete [] m_LastWeights[1];
   if (m_LastWeights[2] != NULL) delete [] m_LastWeights[2];
   if (m_LastBiases[0] != NULL) delete [] m_LastBiases[0];
   if (m_LastBiases[1] != NULL) delete [] m_LastBiases[1];
   if (m_LastBiases[2] != NULL) delete [] m_LastBiases[2];
}

int FIRNet::Load(FILE* InFile)
{
   int i, j;

   // # of epochs trained
   fscanf(InFile, "%i ", &m_EpochsTrained);

   // # of nodes in each layer
   for (i = 0; i < 4; i++)
      fscanf(InFile, "%i ", &m_Nodes[i]);

   // # of taps in each layer
   for (i = 0; i < 3; i++)
      fscanf(InFile, "%i ", &m_Taps[i]);

   // Each weight and bias matrix, and their lengths
   for (i = 0; i < 3; i++)
   {
      fscanf(InFile, "%i", &m_WeightSize[i]);

      if (m_WeightSize[i] == 0)
      {
         m_Weights[i] = NULL;
         m_LastWeights[i] = NULL;
      } else
      {
         m_Weights[i] = new double[m_WeightSize[i]];
         m_LastWeights[i] = new double[m_WeightSize[i]];
         assert(m_Weights[i] != NULL);
         assert(m_LastWeights[i] != NULL);

         for (j = 0; j < m_WeightSize[i]; j++)
            fscanf(InFile, "%lf ", &m_Weights[i][j]);

         ZeroArray(m_LastWeights[i], m_WeightSize[i]);

         m_Biases[i] = new double[m_Nodes[i + 1]];
         m_LastBiases[i] = new double[m_Nodes[i + 1]];
         assert(m_Biases[i] != NULL);
         assert(m_LastBiases[i] != NULL);

         for (j = 0; j < m_Nodes[i + 1]; j++)
            fscanf(InFile, "%lf ", &m_Biases[i][j]);

         ZeroArray(m_LastBiases[i], m_Nodes[i + 1]);
      }
   }

   return 0;
}

int FIRNet::Save(FILE* OutFile)
{
   int i, j;

   // # of epochs trained
   fprintf(OutFile, "%i\n", m_EpochsTrained);

   // # of nodes in each layer
   for (i = 0; i < 4; i++)
      fprintf(OutFile, "%i ", m_Nodes[i]);
   fprintf(OutFile, "\n");

   // # of taps in each layer
   for (i = 0; i < 3; i++)
      fprintf(OutFile, "%i ", m_Taps[i]);
   fprintf(OutFile, "\n");

   // Each weight and bias matrix and their lengths
   for (i = 0; i < 3; i++)
   {
      if (i == 2 && m_Weights[i] == NULL)
         fprintf(OutFile, "0\n");
      else
      {
         fprintf(OutFile, "%i\n", m_WeightSize[i]);
         for (j = 0; j < m_WeightSize[i]; j++)
            fprintf(OutFile, "%g ", m_Weights[i][j]);
         fprintf(OutFile, "\n");

         for (j = 0; j < m_Nodes[i + 1]; j++)
            fprintf(OutFile, "%g ", m_Biases[i][j]);
         fprintf(OutFile, "\n");
      }
   }

   return 0;
}

void FIRNet::Predict(int StartTime, int EndTime, int NumVectors,
   double* InputVectors, double* OutputVectors)
{
   int     i;
   int     MaxEnd;
   int     Dimension = m_Nodes[0];   // Assume input dim = output dim
   double* Prediction;
   double* PredEnd;

   assert(NumVectors > WindowSize());

   printf("Predicting values in time interval [%i, %i]\n", StartTime, EndTime);

   if (EndTime < StartTime)
   {
      i = StartTime;
      StartTime = EndTime;
      EndTime = i;
   }

   assert(StartTime >= 0);
   assert(EndTime >= 0);

   //
   // The time interval of prediction can start and end at any two
   // arbitrary points. What we really do is predict the interval
   // [0, max(EndTime, NumVectors)] and then just take out the
   // subset that we need.
   //

   if (EndTime >= NumVectors)
      MaxEnd = EndTime;
   else
      MaxEnd = NumVectors - 1;

   // Allocate some temporary memory to hold the prediction
   Prediction = new double[(MaxEnd + 1 + WindowSize()) * Dimension];

   for (i = 0; i <= MaxEnd * Dimension; i++)
      Prediction[i] = 0.0;

   // Have the network recreate the original time series
   m_NumInputVectors = NumVectors;
   m_InputVectors = InputVectors;
   // Need to skip first vector
   m_NetOutput = &Prediction[Dimension];
   ForwardPass();    // Feed all input vectors through

   //
   // Use iterative prediction for the rest of the vectors.
   // We start a window of time values at the end of the known
   // vectors; predict the next vector; then move the window one
   // to the right and continue.
   //

   PredEnd = &Prediction[NumVectors * Dimension];

   for (i = 0; i < MaxEnd - NumVectors + 1; i++)
   {
      m_InputVectors = &PredEnd[(i - WindowSize() - 1) * Dimension];
      m_NumInputVectors = WindowSize() + 1;
      m_NetOutput = &PredEnd[(i - WindowSize()) * Dimension];
      ForwardPass();
   }

   // Copy the prediction into the output series
   PredEnd = &Prediction[StartTime * Dimension];
   for (i = 0; i < (EndTime - StartTime + 1) * Dimension; i++)
      OutputVectors[i] = PredEnd[i];

   delete [] Prediction;

   return;
}

void RandomizeArray(double* Array, int Length)
{
   // Seed an entire array with random doubles between 0 and 1
   while (Length-- > 0)
      Array[Length] = (double)rand() / (double)RAND_MAX;
}

void ZeroArray(double* Array, int Length)
{
   while (Length-- > 0)
      Array[Length] = 0.0;
}
