// File:     firnet.h
// Author:   Jeff Miller
// Date:     12-18-97
// Abstract: Class definitions for the FIR neural network engine

#ifndef __FIRNET_H__

#include <stdio.h>

// Transfer function macros
#define HID_FN(x)   (tanh(x))     // Hidden layer transfer function
#define OUT_FN(x)   (x)           // Output layer transfer function
#define D_HID_FN(x) (1.0 - (x*x)) // Derivative of HID_FN w.r.t. x

// Macros for indexing into weight matrices
#define COL_ROW(i, j, ROWS) (i + (j) * (ROWS))
#define COL_ROW_TAP(i, j, tap, ROWS, ROWSXCOLS) (i + (j) * (ROWS) + (tap) * (ROWSXCOLS))

class FIRNet	// A Finite Impulse Response neural network
{
public:

   FIRNet();
   ~FIRNet() { Deallocate(); }

   FIRNet& operator=(FIRNet& net);

   int  Load(FILE* InFile);
   int  Save(FILE* OutFile);

   void SetNodes(int Layer, int NumNodes) {
      m_Nodes[Layer] = NumNodes;	// # of nodes in a layer
   }

   void SetTaps(int Layer, int NumTaps) {
      m_Taps[Layer] = NumTaps;   // # of time taps in a layer
   }

   void SetTrainingParams(int Epochs, double Rate, int Freq, double Momentum, int Decrease) {
      m_NumEpochs = Epochs;
      m_LearningRate = Rate;
      m_DisplayFreq = Freq;      // Display status every X epochs
      m_MomentumRate = Momentum;
      m_DecreaseRate = Decrease; // Linearly decrease learning rate?
   }

   virtual void RandomizeWeights() {}
   virtual int  WindowSize() { return 0; }   // Size of "primer" window
   virtual void Train(int NumVectors, double* InputVectors, double* TargetVectors) {}

   void Predict(int StartTime, int EndTime, int NumVectors, double* InputVectors, double* OutputVectors);

protected:

   void     Deallocate();

   //
   // State
   //

   int      m_Nodes[4];          // # of nodes in input/hidden/output layers
   int      m_Taps[3];         	// # of taps in input/hidden layers

   double*  m_Weights[3];        // Intra-layer weight matrices
   int      m_WeightSize[3];     // # elements in each weight matrix
   double*  m_Biases[3];         // Bias vectors for each layer

   double*  m_LastWeights[3];    // Last weight matrix for momentum learning
   double*  m_LastBiases[3];     // Last bias matrix for momentum learning

   //
   // Training
   //

   int      m_NumEpochs;         // # of epochs to train for
   int      m_EpochsTrained;     // Total epochs this net has trained
   double   m_LearningRate;      // Weight adjustment learning rate
   int      m_DisplayFreq;       // Display status every X epochs
   double   m_MomentumRate;      // Momentum weighting constant
   int      m_DecreaseRate;      // Flag to decrease rate during training

   double*  m_InputVectors;      // Input vectors to train on
   double*  m_TargetVectors;     // Desired output vectors
   int      m_NumInputVectors;	// Total # of input vectors

   int      m_Mask;              // Circular index
   double*  m_MSE;               // Mean squared error per epoch
   double*  m_CSE;               // Cumulative squared error
   double*  m_NetOutput;         // Network outputs
   double** m_NetDelta;          // Network error
   double** m_HiddenOutput;      // Hidden layer outputs
   double*  m_HiddenDelta;       // Hidden layer error

   virtual void ForwardPass() {} // Implemented by specific FIR classes
};

class FIRNet2Layer : public FIRNet	// Special case: 2-layer FIR
{
public:

   virtual void RandomizeWeights();

   virtual int WindowSize() {
      return (m_Taps[0] + m_Taps[1] - 2);
   }

   virtual void Train(int NumVectors, double* InputVectors, double* TargetVectors);

protected:

   void FwdPassLayer1(int Index);
   void FwdPassLayer2(int Index);
   void BackPropLayer1(int Index);
   void BackPropLayer2(int Index);
   void UpdateWeights(int Index);
   void CalcMSE(int Epoch);
   void ResetDeltas();

   virtual void ForwardPass();
};

class FIRNet3Layer : public FIRNet	// Special case: 3-layer FIR
{
public:

   virtual void RandomizeWeights();

   virtual int WindowSize() {
      return (m_Taps[0] + m_Taps[1] + m_Taps[2] - 3);
   }

   virtual void Train(int NumVectors, double* InputVectors, double* TargetVectors);

protected:

   double** m_HiddenOutput2;	// 2nd hidden layer output
   int      m_Mask2;          // Extra circular index
   double** m_HiddenDelta2;   // 2nd hidden layer delta

   void FwdPassLayer1(int Index);
   void FwdPassLayer2(int Index);
   void FwdPassLayer3(int Index);
   void BackPropLayer1(int Index);
   void BackPropLayer2(int Index);
   void BackPropLayer3(int Index);
   void UpdateWeights(int Index);
   void CalcMSE(int Epoch);
   void ResetDeltas();

   virtual void ForwardPass();
};

void RandomizeArray(double* Array, int Length);
void ZeroArray(double* Array, int Length);

#define __FIRNET_H__
#endif
