// File:     main.cc
// Author:   Jeff Miller
// Date:     12-17-97
// Abstract: Command line interface to FIR time series prediction engine

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include "series.h"
#include "firnet.h"

const char* Usage =
   "Usage: firnet -f filename [switches...]\n"
   "\n"
   "Switches:\n"
   "    -l layers      # of layers in the FIR network: 2 (default) or 3\n"
   "    -n n1 [n2]     # of nodes in each hidden layer\n"
   "    -t n1 n2 [n3]  # of taps in each layer\n"
   "    -p start end   Range of time series values to predict\n"
   "    -e epochs      # of epochs to train for\n"
   "    -c             Force training (if disabled by loading state)\n"
   "    -r rate        Learning rate\n"
   "    -m alpha       Use momentum with parameter alpha in [0, 1]\n"
   "    -d             Linearly decrease learning rate";

int main(int argc, char** argv)
{
   TimeSeries T, P;
   char*      FileName = NULL;
   int        Layers = 2;
   int        StartTime = 0;
   int        EndTime = -1;
   int        Epochs = 75;
   int        DispFreq = 5;
   double     Rate = 0.005;
   double     Alpha = 0.0;
   int        i, j;
   int        ForceTrain = 0;
   int        WeightsLoaded = 0;
   int        DecreaseRate = 0;

   srand(time(NULL));	// Seed random number generator

   // Convert all switches to lower case
   for (i = 1; i < argc; i++)
   {
      if (argv[i][0] == '-')
         for (j = 0; j < strlen(argv[i]); j++)
            argv[i][j] = tolower(argv[i][j]);
   }

   // Seek out the -layers switch on the command line first
   for (i = 1; i < argc; i++)
   {
      if (strcmp(argv[i], "-l") == 0)
      {
         if (i + 1 >= argc)
         {
            printf("Expected parameter after %s\n", argv[i]);
            return 1;
         }
         Layers = atoi(argv[++i]);
      }
   }

   assert(Layers == 2 || Layers == 3);

   // Some default network parameters
   T.SetNetworkSize(Layers);
   T.SetHiddenNodes(0, 5);
   T.SetHiddenNodes(1, 4);
   T.SetLayerTaps(0, 4);
   T.SetLayerTaps(1, 3);
   T.SetLayerTaps(2, 2);

   // Parse the command line to change program parameters
   for (i = 1; i < argc; i++)
   {
      // Command line help
      if (strcmp(argv[i], "-h") == 0)
      {
         printf("%s\n", Usage);
         return 1;
      } else
      // # of epochs to train for
      if (strcmp(argv[i], "-e") == 0)
      {
         if (i + 1 >= argc)
         {
            printf("Expected parameter after %s\n", argv[i]);
            return 1;
         }
         Epochs = atoi(argv[++i]);
      } else
      // # of layers in network (2 or 3)
      if (strcmp(argv[i], "-l") == 0)
         Layers = atoi(argv[++i]);
      else
      // # of nodes in each layer -- sequence of ints follows
      if (strcmp(argv[i], "-n") == 0)
      {
         if (i + 1 >= argc)
         {
            printf("Expected parameter after %s\n", argv[i]);
            return 1;
         }
         T.SetHiddenNodes(0, atoi(argv[++i]));
         if (i + 1 < argc && atoi(argv[i + 1]) > 0)
            T.SetHiddenNodes(1, atoi(argv[++i]));
      } else
      // # of taps in each layer -- sequence of ints follows
      if (strcmp(argv[i], "-t") == 0)
      {
         if (i + 2 >= argc)
         {
            printf("Expected parameters after %s\n", argv[i]);
            return 1;
         }
         T.SetLayerTaps(0, atoi(argv[++i]));
         if (i + 1 < argc && atoi(argv[i + 1]) > 0)
         {
            T.SetLayerTaps(1, atoi(argv[++i]));
            if (i + 1 < argc && atoi(argv[i + 1]) > 0)
               T.SetLayerTaps(2, atoi(argv[++i]));
         }
      } else
      // Range of time values to predict
      if (strcmp(argv[i], "-p") == 0)
      {
         if (i + 2 >= argc)
         {
            printf("Expected parameters after %s\n", argv[i]);
            return 1;
         }
         StartTime = atoi(argv[++i]);
         EndTime = atoi(argv[++i]);
      } else
      // Input series file name
      if (strcmp(argv[i], "-f") == 0)
      {
         if (i + 1 >= argc)
         {
            printf("Expected parameter after %s\n", argv[i]);
            return 1;
         }
         FileName = argv[++i];
      } else
      // Force training
      if (strcmp(argv[i], "-c") == 0)
         ForceTrain = 1;
      else
      // Learning rate
      if (strcmp(argv[i], "-r") == 0)
      {
         if (i + 1 >= argc)
         {
            printf("Expected parameter after %s\n", argv[i]);
            return 1;
         }
         Rate = atof(argv[++i]);
      } else
      // Use momentum
      if (strcmp(argv[i], "-m") == 0)
      {
         if (i + 1 >= argc)
         {
            printf("Expected parameter after %s\n", argv[i]);
            return 1;
         }
         Alpha = atof(argv[++i]);
      } else
      // Decrease learning rate during training
      if (strcmp(argv[i], "-d") == 0)
         DecreaseRate = 1;
      else
      // Unknown switch
      {
         printf("Unknown switch %s\n", argv[i]);
         return 1;
      }
   }

   if (FileName == NULL)
   {
      printf("%s\n", Usage);
      return 1;
   }

   if (T.LoadSeries(FileName) != 0)
   {
      printf("There was an error reading the input file.\n");
      return 1;
   }

   if (T.LoadNetwork(FileName) == 0)
      WeightsLoaded = 1;
   else
      T.RandomizeNetwork();

   T.SetTrainingParams(Epochs, Rate, DispFreq, Alpha, DecreaseRate);

   if (!WeightsLoaded || ForceTrain)
      T.TrainNetwork();

   // If the prediction range was not set by the user, just do 10%
   if (EndTime == -1)
      EndTime = T.Length() - 1 + T.Length()/10;

   T.Predict(StartTime, EndTime, P);

   P.SaveSeries(FileName);
   T.SaveNetwork(FileName);

   return 0;
}
