// File:     series.cc
// Author:   Jeff Miller
// Date:     12-17-97
// Abstract: Implementation of time series prediction class

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include "series.h"

TimeSeries::TimeSeries()
{
   m_Vectors = NULL;
   m_NumVectors = 0;
   m_NetworkSize = 2;
   m_Network = new FIRNet2Layer;
}

void TimeSeries::Deallocate()
{
   if (m_Vectors != NULL)
   {
      delete [] m_Vectors;
      m_Vectors = NULL;
   }

   if (m_Network != NULL)
   {
      delete m_Network;
      m_Network = NULL;
   }
}

void TimeSeries::PrintSeries()
{
   int     i, j;
   double* p;

   for (i = 0, p = m_Vectors; i < m_NumVectors; i++)
   {
      for (j = 0; j < m_Dimension; j++)
         printf("%lf ", *p++);
      printf("\n");
   }
}

int TimeSeries::LoadSeries(char* FileName)
{
   FILE*   InFile;
   char    Line[512];
   char*   p;
   int     n, j;
   double* Value;

   if ((InFile = fopen(FileName, "r")) == NULL)
      return 1;	// Unable to open file

   //
   // Count the number of input dimensions by examining the first line.
   //

   if (fgets(Line, 512, InFile) == NULL || Line[0] == 0)
      return 2;	// Unable to read from file

   for (p = Line, j = m_Dimension = 0; *p != 0; p++)
   {
      if (j == 0 && !isspace(*p))
      {
         j = 1;
         m_Dimension++;
      } else
      if (j == 1 && isspace(*p))
         j = 0;
   }

   //
   // Get an upper bound on the number of points in the series
   // and allocate the necessary memory.
   //

   j = 1;
   while (fgets(Line, 512, InFile) != NULL)
      j++;

   if (m_Vectors != NULL)
      delete [] m_Vectors;

   m_Vectors = new double[j * m_Dimension + 1];

   //
   // Now go back and read in the file.
   //

   fseek(InFile, 0, SEEK_SET);
   m_NumVectors = 0;

   for (n = 0, j *= m_Dimension, Value = m_Vectors;
      fscanf(InFile, "%lf ", Value) != EOF && j >= 0;
      Value++, n++, j--);

   if (j == -1)
      return 4;	// Went over upper bound on storage; file is not uniform

   if ((n % m_Dimension) != 0)	// Make sure the series is uniform
      return 3;	// Series is not uniform

   m_NumVectors = n / m_Dimension;

   printf("*** Read %i R^%i vectors from %s\n", m_NumVectors, m_Dimension, FileName);

   fclose(InFile);

   // Set the number of nodes in the input and output layers
   m_Network->SetNodes(0, m_Dimension);
   m_Network->SetNodes(m_NetworkSize, m_Dimension);

   return 0;	// No error
}

int TimeSeries::SaveSeries(char* BaseName)
{
   int     i, j;
   double* p;
   FILE*   OutFile;

   if (strchr(BaseName, '.') != NULL)
      strcpy(strchr(BaseName, '.'), ".out");
   else
      strcat(BaseName, ".out");

   if ((OutFile = fopen(BaseName, "wb")) == NULL)
      return 1;

   for (i = 0, p = m_Vectors; i < m_NumVectors; i++)
   {
      for (j = 0; j < m_Dimension; j++)
         fprintf(OutFile, "%lf ", *p++);
      fprintf(OutFile, "\n");
   }

   fclose(OutFile);

   printf("*** Saved %i vectors to %s\n", m_NumVectors, BaseName);

   return 0;
}

int TimeSeries::LoadNetwork(char* BaseName)
{
   FILE* InFile;
   int   r;

   if (strchr(BaseName, '.') != NULL)
      strcpy(strchr(BaseName, '.'), ".net");
   else
      strcat(BaseName, ".net");

   if ((InFile = fopen(BaseName, "r")) == NULL)
      return 1;

   // Read in the # of layers in the network
   fscanf(InFile, "%i", &r);
   assert(r == 2 || r == 3);
   SetNetworkSize(r);

   r = m_Network->Load(InFile);

   fclose(InFile);

   printf("*** Read network state from %s\n", BaseName);

   return r;
}

int TimeSeries::SaveNetwork(char* BaseName)
{
   FILE* OutFile;
   int   r;

   if (strchr(BaseName, '.') != NULL)
      strcpy(strchr(BaseName, '.'), ".net");
   else
      strcat(BaseName, ".net");

   if ((OutFile = fopen(BaseName, "w")) == NULL)
      return 1;

   fprintf(OutFile, "%i\n", m_NetworkSize);

   r = m_Network->Save(OutFile);

   fclose(OutFile);

   printf("*** Saved network state to %s\n", BaseName);

   return r;
}

void TimeSeries::BlankSeries(int Dimension, int NumVectors)
{
   int     i;
   double* p;

   Deallocate();

   m_Dimension = Dimension;
   m_NumVectors = NumVectors;

   m_Vectors = new double[Dimension * NumVectors];

   for (i = m_Dimension * m_NumVectors, p = m_Vectors; i > 0; i--)
      *p++ = 0.0;
}

void TimeSeries::SetNetworkSize(int Size)
{
   FIRNet* NewNet;

   assert(Size == 2 || Size == 3);
   m_NetworkSize = Size;		// Size = 2 or 3

   if (Size == 2)
      NewNet = new FIRNet2Layer;
   else
      NewNet = new FIRNet3Layer;
   assert(NewNet != NULL);

   *NewNet = *m_Network;
   delete m_Network;
   m_Network = NewNet;
}

void TimeSeries::Predict(int StartTime, int EndTime, TimeSeries& Output)
{
   assert(StartTime >= 0);
   assert(StartTime <= EndTime);

   // Clear the output time series
   Output.BlankSeries(m_Dimension, EndTime - StartTime + 1);
   // Use iterative prediction
   m_Network->Predict(StartTime, EndTime, m_NumVectors, m_Vectors, &Output[0]);
}
